/***  fft.c
*
*	one-dimensional Fast Fourier Transform
*
*            -1   Inverse Transform
*       mode  0   Setup Tables
*             1   Forward Transform
*
*       INV is bit-reversal table
*
****/

#include <stdio.h>
#include <math.h>

void fft(float real[], float imag[], int n, int mode);

static int m,nm1,nv2,inv[512];
static float cosine[16],sine[16];

#ifdef TEST

DPRINT(FILE *out, float *f, int n)
{
	int i;
	float *z;
	fprintf(out," N      Value\n");
	if (out!=stdout) printf(" N       Value\n");
	for (i = 0, z=f; i<n; i++,z++) {
		fprintf(out,"%2d %12.4f\n",i,*z);
		if (out!=stdout) printf("%2d %12.4f\n",i,*z);
	}
}

#include <malloc.h>

main()
{
	float *f,*g,*z,tmp;
	FILE *out;
	int i,na,width;

	out = fopen("fft.dat","wt");

	printf("\nEnter cell width:"); scanf("%d",&na);
	printf("\nEnter width of rect:");scanf("%d",&width);

	f = (float *)malloc(na*sizeof(float));
	g = (float *)malloc(na*sizeof(float));

	printf("\nSetup procedure:\n\n");
	fft(f,g,na,0);
	
	for (i = 0; i< na; i++) {
		f[i] = 0.0;
		g[i] = 0.0;
	}
	tmp = 0.5/width;
	f[0] = tmp;

	for (i = 1; i<=width; i++) {
		if (i == width) tmp /= 2.0;
		f[i] = tmp;
		f[na-i] = tmp;
	}

	DPRINT(out,f,na);

/* Fourier series should be sinc function */

	fft(f,g,na,1); /*Forward*/
	DPRINT(out,f,na);

/* If we square it, we should get sinc**2 */

	for (i = 0; i<na; i++) {
	tmp = f[i];
	f[i] = tmp*tmp;
	g[i] = 0.0;
}

/* Reverse transform gives us triangle function */

	fft(f,g,na,-1); /* Reverse */
	DPRINT(out,f,na);
}
#endif

#define SWAP(a,b) { t = (a); (a) = (b); (b) = t; }

void fft(float real[], float imag[], int n, int mode)
{
	int i, ip, j, k, l, lv2;
	double ang;
	float t, tmp;
	float treal, timag, ureal, uimag, vreal, vimag, wreal, wimag;

	if (mode==0) {

		/*  Setup Tables */
	
		/* find m such that n = 2^m */
	
		for (j=1,m=0; j<n; j<<=1, m++);
		if (j!=n) {
			fputs("array must be power of two",stderr);
			return;
		}
		nv2 = n/2;
		nm1 = n-1;
		/*  inv is the bit-reversal table */

		for (i=0, j=0; i < nm1; i++) {
			inv[i] = j;
			k = nv2;
			while (k<=j) {
				j -=  k;
				k >>= 1;
			}
			j += k;
		}
		inv[nm1]=nm1;
		
		/* tabl contains necessary trig functions */

		cosine[0] = -1.0; sine[0] = 0.0;
		ang = 3.14159265359;
		for (k=1; k < m; k++ ) {
			ang /= 2.;
			cosine[k] = cos(ang);
			sine[k] = sin(ang);
		}
		return;
	}

	/*  Forward Transform */

	for (i=0; i < n; i++ ) {
		j = inv[i];
		if ( i < j ) {
			SWAP(real[i],real[j]);
			SWAP(imag[i],imag[j]);
		}
	}

	l = 1;
	for (k=0; k < m; k++ ){
		lv2 = l;
		l <<= 1;
		ureal = 1.0;  uimag = 0.0;
		wreal = cosine[k]; wimag = sine[k];
		for (j=0; j < lv2; j++ ) {
			for (i=j; i<n; i+=l) {
				ip = i + lv2;
				vreal = real[ip]; vimag = imag[ip];
				treal = ureal*vreal-uimag*vimag;
				timag = ureal*vimag+uimag*vreal;
				real[ip] = real[i] - treal;
				imag[ip] = imag[i] - timag;
				real[i] += treal; imag[i] += timag;
			}
			/* u = u*w */
			tmp = ureal*wimag+uimag*wreal;
			ureal = ureal*wreal-uimag*wimag;
			uimag = tmp;
		}
	}

	if (mode > 0 )	return;

	/*  Inverse Transform */

	for (i=1, ip=nm1; i<nv2; i++,ip-- ) {
		SWAP(real[i],real[ip]);
		SWAP(imag[i],imag[ip]);
	}

	for (i=0; i < n; i++ ) {
		real[i] /= n;
		imag[i] /= n;
	}
}
